@@ -18,17 +18,14 @@
  * Author: David Zeuthen <davidz@redhat.com>
  */
 
-#ifndef __GOA_PANEL_H__
-#define __GOA_PANEL_H__
+#pragma once
 
 #include <shell/cc-panel.h>
 
 G_BEGIN_DECLS
 
-#define CC_TYPE_GOA_PANEL (cc_goa_panel_get_type ())
+#define CC_TYPE_ONLINE_ACCOUTS_PANEL (cc_online_accounts_panel_get_type ())
 
-G_DECLARE_FINAL_TYPE (CcGoaPanel, cc_goa_panel, CC, GOA_PANEL, CcPanel)
+G_DECLARE_FINAL_TYPE (CcOnlineAccountsPanel, cc_online_accounts_panel, CC, ONLINE_ACCOUNTS_PANEL, CcPanel)
 
 G_END_DECLS
-
-#endif /* __GOA_PANEL_H__ */
